;; Generated package description from el-search.el
(define-package "el-search" "0.2.3" "Expression based incremental search for emacs-lisp-mode" '((emacs "25")) :url "http://elpa.gnu.org/packages/el-search.html" :keywords '("lisp"))
